#!!

Create full swing bar from swing
================================

Input:
Swing - list<pos, value>
Beats - list<pos, barnum, beatnum>

Output:
Swing - list<pos, value> ;; All values are filled out.


Create swing filledout list
===========================

Input:
======

PrevSwing - list<pos, value> ;; Swing of previous track. If a value is 0, it means that automatic repeat of swing stops. (Same thing also happens if the measure changes)
                             ;; Perhaps BlockSwing is better than PrevTrackSwing? (no, because then we would need a custom "BlockSwing" track). 
                             ;; But perhaps FirstTrackSwing instead of PrevTrackSwing?
                             ;;    + If adding swing at Track #2, there won't be any swing on Track #3, unless there is swing on Track #1
                             ;;    - Track #1 will have a special meaning, but maybe that is not such a bad thing.
                             ;;    - Prev track could give some interesting results.
                             ;;    + If adding a custom BlockSwing track later, the behavior will be more similar and it might be easier to load older songs.
                             ;;    - If using previous track, it doesn't make very much sense to make a custom BlockSwing track.
                             ;;    - It might be more common that you want to use previous track.
                             ;;    + It might be more common to use a default swing track.
                             ;;    + Probably less confusing, but not sure. For sure less confusing to use custom BlockSwing track though.
                             ;;    + A BlockSwing track is more excplicit (not sure)
                             ;;    - The missing numbers are filled in automatically anyway, so it doesn't matter that much if it's less confusing.
                             ;;    + If changing swing of track 3, it's likely that you don't want to change swing of track 4.
Swing     - list<pos, value> ;; Same behavior if value is 0 as above.
Beats     - list<pos, barnum, beatnum>
LPBs      - list<pos, lpb>
DefaultLPB - integer
NumLines  - integer

Output:
list<pos, value> ;; A value can not be 0, and all automatic repeat of swing has been filled out.


!!#

::::::::::::::::::::::::::::::::::::::
;;;;;;;;;;;; BAR :::::::::::::::::::::
::::::::::::::::::::::::::::::::::::::

(define (create-beat line num)
  (list line num))

(define (get-beat-line beat)
  (car beat))

(define (get-beat-num beat)
  (cadr beat))

(define-struct bar
  :barnum
  :beats)

(define (get-first-line-in-bar bar)
  (get-beat-line 
   (car (bar :beats))))




::::::::::::::::::::::::::::::::::::::
;;;;;;;;;;;; SWING :::::::::::::::::::
::::::::::::::::::::::::::::::::::::::

(define (create-swing line weight)
  (list line weight))

(define (get-swing-line swing)
  (car swing))

(define (get-swing-weight swing)
  (cadr swing))


(define-struct bar-swing
  :barnum
  :swings
  :num-lines)


;;(define (bars-have-same-signatures? prev-bar prev-bar-length bar curr-bar-length)
;;  (= prev-bar-length curr-bar-length)) ;; Probably good enough.


(define *test-shuffle-swing* '((0 1)
                               (2 2)
                               (4 1)
                               (6 2)
                               (8 1)
                               (10 2)
                               (12 1)
                               (14 2)))

(define *test-4/4-beats* (list (make-bar :barnum 0 :beats '(( 0 0)
                                                            ( 4 1)
                                                            ( 8 2)
                                                            (12 3)))
                               (make-bar :barnum 1 :beats '((16 0)
                                                            (20 1)
                                                            (24 2)
                                                            (28 3)))))


#||
;; Currently just moves the first swing in the bar to the start of the bar. (not needed, incorportated into 'group-swing-by-bar' instead)
(define (add-bar-start-to-swing bars swing)
  (if (null? bars)
      '()
      (let* ((bar (car bars))
             (next-bar (cl-cadr bars))
             (first-line (car bar))
             (next-line (or (and next-bar (car (cadr next-bar)))
                            num-lines)))

        )))

(***assert*** (add-bar-start-swings '((0 (0 0))
                                      (1 (2 0)))
                                    '((0 1)
                                      (1 2)
                                      (3 5)))
              '((0 1)
                (1 2)
                (2 5)))
(***assert*** (add-bar-start-swings '((0 (0 0))
                                      (1 (2 0)))
                                    '((0 1)
                                      (1 2)
                                      (3 5)
                                      (7/2 6)))
              '((0 1)
                (1 2)
                (2 5)
                (7/2 6)))
||#           

(define (group-swing-by-bar bars swings num-lines)
  (if (null? bars)
      '()
      (let* ((bar (car bars))
             (next-bar (begin (c-display ":bar" bar ":bars" bars) (cl-cadr bars)))
             (barnum (bar :barnum))
             (first-line (get-first-line-in-bar bar))
             (next-line (if next-bar
                            (get-first-line-in-bar next-bar)
                            num-lines)))
        (let loop ((swings swings)
                   (result '())
                   (is-first-bar #t))
          (define (gotit)
            ;;(c-display "gotit" first-line result)
            (if (null? result)
                (group-swing-by-bar (cdr bars) swings num-lines)
                (cons (make-bar-swing :barnum barnum
                                      :swings (map (lambda (swing)
                                                     (create-swing (- (get-swing-line swing)
                                                                      first-line)
                                                                   (get-swing-weight swing)))
                                                   (reverse result))
                                      :num-lines (- next-line first-line))
                      (group-swing-by-bar (cdr bars) swings num-lines))))
          ;;(c-display "swing" swing)
          (c-display ":bar" bar (cl-cadr next-bar) ":first-line" first-line ":swings" swings ":result" result ":next-line" next-line)
          (if (null? swings)
              (gotit)
              (let* ((swing (car swings))
                     (swing-line (get-swing-line swing)))
                (if (>= swing-line next-line)
                    (gotit)
                    (loop (cdr swings)
                          (cons (if (and is-first-bar
                                         (> swing-line first-line))
                                    (create-swing first-line
                                                  (get-swing-weight swing))
                                    swing)
                                result)
                          #f))))))))
        
(pretty-print (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 3 5))
                                  4))
        

(***assert*** (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 2 4)
                                        (create-swing 3 5))
                                  4)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 4)
                                                            (create-swing 1 5))
                                    :num-lines 2)))


(***assert*** (group-swing-by-bar (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                        (make-bar :barnum 1 :beats (list (create-beat 2 0))))
                                  (list (create-swing 0 1)
                                        (create-swing 1 2)
                                        (create-swing 3 5))
                                  4)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 5))
                                    :num-lines 2)))


#||
(***assert*** (group-swing-by-bar *test-4/4-beats*
                                  *test-shuffle-swing*
                                  64)
              `((0 ,*test-shuffle-swing*)))
||#


;; Add autoinserted swings.
(define (create-filledout-swings bars global-swings track-swings num-lines)  
  (let loop ((bars bars)
             (prev-bar #f)
             (curr-swing #f)
             (global-swings global-swings) ;;(group-swing-by-bar bars global-swing num-lines))
             (track-swings track-swings)) ;;(group-swing-by-bar bars track-swing num-lines)))
    (cond ((null? bars)
           (assert (null? global-swings))
           (assert (null? track-swings))
           '())

          ((and (not curr-swing)
                (null? global-swings)
                (null? track-swings))
           '())

          (else

           (define bar (car bars))
           (define barnum (bar :barnum))

           (while (and (not (null? global-swings))
                       (< ((car global-swings) :barnum) barnum))
             (set! global-swings (cdr global-swings)))

           (define global-swing (cl-car global-swings))
           (define track-swing (cl-car track-swings))

           (cond ((and track-swing
                       (= barnum (track-swing :barnum)))
                  (cons track-swing
                        (loop (cdr bars)
                              bar
                              track-swing
                              global-swings
                              (cdr track-swings))))

                 ((and global-swing
                       (= barnum (global-swing :barnum)))
                  (cons global-swing
                        (loop (cdr bars)
                              bar
                              global-swing
                              (cdr global-swings)
                              track-swings)))

                 ((and curr-swing
                       (let* ((curr-line (get-first-line-in-bar bar))
                              (next-bar (cl-cadr bars))
                              (next-line (if next-bar
                                             (get-first-line-in-bar next-bar)
                                             num-lines))
                              (curr-bar-length (- next-line
                                                  curr-line)))
                         (= (curr-swing :num-lines)
                            curr-bar-length)))

                  (cons (copy-bar-swing curr-swing :barnum barnum)
                        (loop (cdr bars)
                              bar
                              curr-swing
                              global-swings
                              track-swings)))

                 (else
                  (loop (cdr bars)
                        bar
                        #f
                        global-swings
                        track-swings)))))))



(***assert*** (create-filledout-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                             (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                             (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                       (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                     (create-swing 1 2))
                                                             :num-lines 2))
                                       (list (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                     (create-swing 1 1))
                                                             :num-lines 2))
                                       6)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                            (create-swing 1 1))
                                    :num-lines 2)))

;; Test that we don't repeat swing when there is a signature change. (actually, when the bar changes number of lines)
(***assert*** (create-filledout-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                             (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                             (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                       (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                     (create-swing 1 2))
                                                             :num-lines 2))
                                       (list (make-bar-swing :barnum 1 :swings (list (create-swing 0 2)
                                                                                     (create-swing 1 1))
                                                             :num-lines 2))
                                       5)
              (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                            (create-swing 1 2))
                                    :num-lines 2)
                    (make-bar-swing :barnum 1 :swings (list (create-swing 0 2)
                                                            (create-swing 1 1))
                                    :num-lines 2)))


#||

Create swing multiplication list
================================

Input:
=======
FilledoutSwing - list<pos, value>
Beats - list<pos, barnum, beatnum>


Output:
list<pos,tempo_multiplication,hold>

||#

(define-struct tempo-multiplier
  :line
  :multiplier
  :hold)

(define (create-tempo-multipliers-from-swing start-line bar-swing)
  (define swings (bar-swing :swings))
  (define num-lines (bar-swing :num-lines))
  (define end-line (+ start-line num-lines))
  (define weight-sum (let loop ((swings swings))
                       (if (null? swings)
                           0
                           (let* ((swing (car swings))
                                  (line1 (get-swing-line swing))
                                  (next-swing (cl-cadr swings))
                                  (line2 (if next-swing
                                             (get-swing-line next-swing)
                                             num-lines))
                                  (duration (- line2 line1)))
                             (+ (* duration 
                                   (/ 1 (get-swing-weight swing)))
                                (loop (cdr swings)))))))
  (let loop ((swings swings))
    (if (null? swings)
        '()
        (let* ((swing (car swings))
               (line1 (get-swing-line swing))
               (next-swing (cl-cadr swings))
               (line2 (if next-swing
                          (get-swing-line next-swing)
                          num-lines))
               (duration (- line2 line1)))
          (cons (make-tempo-multiplier :line (+ start-line line1)
                                       :multiplier (/ (* duration
                                                         (/ 1 (get-swing-weight swing))
                                                         num-lines)
                                                      weight-sum)
                                       :hold #t)
                (loop (cdr swings)))))))

(***assert*** (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                              (create-swing 1 2))
                                                                      :num-lines 2))
              (list (make-tempo-multiplier :line 10 :multiplier 4/3 :hold #t)
                    (make-tempo-multiplier :line 11 :multiplier 2/3 :hold #t)))
              
(***assert*** (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                              (create-swing 1 2))
                                                                      :num-lines 3))
              (list (make-tempo-multiplier :line 10 :multiplier 3/2 :hold #t)
                    (make-tempo-multiplier :line 11 :multiplier 6/4 :hold #t)))
              

#!!
(pp (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2))
                                                            :num-lines 3)))
(pp (create-tempo-multipliers-from-swing 10 (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2)
                                                                                    (create-swing 2 2))
                                                            :num-lines 3)))
!!#


(define (create-tempo-multipliers-from-swings bars bar-swings)
  (define bar (cl-car bars))
  (define bar-swing (cl-car bar-swings))
  (define line (and bar (get-first-line-in-bar bar)))

  (if (not bar-swing)
      (if bar
          (list (make-tempo-multiplier :line line :multiplier 1 :hold 3t))
          '())
      (append (create-tempo-multipliers-from-swing line bar-swing)
              (create-tempo-multipliers-from-swings (cdr bars)
                                                    (cdr bar-swings)))))

#!!
(create-tempo-multipliers-from-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                            (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                            (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                      (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2))
                                                            :num-lines 2)
                                            (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                    (create-swing 1 2))
                                                            :num-lines 2)
                                            (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                    (create-swing 1 1))
                                                            :num-lines 2)))
!!#

(***assert*** (create-tempo-multipliers-from-swings (list (make-bar :barnum 0 :beats (list (create-beat 0 0)))
                                                          (make-bar :barnum 1 :beats (list (create-beat 2 0)))
                                                          (make-bar :barnum 2 :beats (list (create-beat 4 0))))
                                                    (list (make-bar-swing :barnum 0 :swings (list (create-swing 0 1)
                                                                                                  (create-swing 1 2))
                                                                          :num-lines 2)
                                                          (make-bar-swing :barnum 1 :swings (list (create-swing 0 1)
                                                                                                  (create-swing 1 2))
                                                                          :num-lines 2)
                                                          (make-bar-swing :barnum 2 :swings (list (create-swing 0 2)
                                                                                                  (create-swing 1 1))
                                                                          :num-lines 2)))
              (list (make-tempo-multiplier :line 0 :multiplier 4/3 :hold #t)
                    (make-tempo-multiplier :line 1 :multiplier 2/3 :hold #t)
                    (make-tempo-multiplier :line 2 :multiplier 4/3 :hold #t)
                    (make-tempo-multiplier :line 3 :multiplier 2/3 :hold #t)
                    (make-tempo-multiplier :line 4 :multiplier 2/3 :hold #t)
                    (make-tempo-multiplier :line 5 :multiplier 4/3 :hold #t)))
                    


Input:
=======

DefaultLPB - integer
DefaultBPM - integer
LPB - list<pos,value,is_hold>
BPM - list<pos,value,is_hold>
TempoAutomation - list<pos,tempoMul>, with at least two nodes. The first node has position 0, the last node has position NumLines.
Swing - list<?>
<strike>NumLines - integer</strike>
<strike>Samplerate - integer</strike>


Output:
=======

<strike>Timing - list, with at least one node per integer. First node has the value 0. Length is NumLines + 1.</strike>
Tempo automation: list<pos,tempo multiplication, is_hold>


Swing:
=======

0 -> 1
1 -> 2
2 -> 1
3 -> 2
4 -> 1
5 -> 2
6 -> 1
 ....
 ===>
0 -> 1.33
1 -> 0.66
2 -> 1.33
3 -> 0.66
 ....


LPB:
======
0 -> 1
1 -> 2
2 -> 1
3 -> 2
4 -> 1
5 -> 2
6 -> 1
 ....
 ===>
0 -> 1   (1/1
1 -> 0.5 (1/2)
2 -> 1   (1/2)
3 -> 0.5 (1/1
 ....


BPM:
======
0 -> 100
1 -> 200
2 -> 100
3 -> 200
4 -> 100
5 -> 200
6 -> 100
 ....
 ===>
0 -> 60/100
1 -> 60/200
2 -> 60/100
3 -> 60/200
 ....



(define (get-duration duration-in-lines samplerate lpb bpm tempo-multiplier)
  (/ (* duration-in-lines 60 samplerate)
     (* lpb bpm tempo-multiplier)))
  

(***assert*** (/ (get-duration 64 48000 4 120 1)
                 48000)
              8)

(define (create-posval pos val)
  (cons pos val))

(define (posval-pos posval)
  (car posval))

(define (posval-val posval)
  (cdr posval))

;; Could we add swing in the same type of format as lpb and bpm?
;; In case, maybe we could merge lpb/bpm/swing before calling this function.
(define (create-timing num-lines samplerate default-lpb lpbs default-bpm bpms)
  (let loop ((pos 0)
             (lpb default-lpb)
             (lpbs lpbs)
             (bpm default-bpm)
             (bpms bpms))
    (define lpb1 (cl-car lpbs))
    (define bpm1 (cl-car bpm1))
    (define next-line (1+ (floor pos)))

    (cond ((> pos num-lines)
           (assert (null? lpbs))
           (assert (null? bpms))
           '())

          ;; something like this.
          ((and lpb (<= (posval-pos lpb1)
                        next-line))
           (define lpb1pos (posval-pos lpb1))
           (cons (+ last-time
                    (get-duration (- lpb1pos last-pos)
                                  samplerate
                                  lpb
                                  pbm
                                  1.0))
                 (loop lpb1pos
                       (posval-value lpb1)
                       (cdr lpbs)
                       bpm
                       bpms)))

          (else
           (cons (get-duration line samplerate default-lpb default-bpm 1.0)
                 (loop next-line
                       lpbs
                       bpms))))))


(let ((num-lines 2)
      (samplerate 48000)
      (default-lpb 4)
      (lpbs '())
      (default-bpm 125)
      (bpms '()))
  (***assert*** (create-timing num-lines samplerate default-lpb lpbs default-bpm bpms)
                (list 0
                      (get-duration 1 samplerate default-lpb default-bpm 1)
                      (get-duration 2 samplerate default-lpb default-bpm 1))))

(/ (get-time 64 48000 4 120 .5)
   48000.0)

